/* 1541, Thu 17 Mar 11 (PDT)
   1628, Sat 17 Dec 10 (NZDT) 
   0927, Mon 30 Aug 10 (NZST)
   1550, Tue 27 Jul 10 (CEST)
   1253, Wed 19 May 10 (NZST)
   1955, Thu 22 Apr 10 (PDT)
   1445, Wed  5 Dec 07 (NZDT)

   owt.h:  Globals for OneWayTraffic Monitor

   iatmon/owtmon: One-Way Traffic Monitor testbed
   Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sys/types.h>
#include <sys/stat.h>

#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <syslog.h>
#include <unistd.h>

#include <pthread.h>
#include "semaphore.h"

#include "libtrace.h"

#include "../owt_conf.h"

/* odin:      DFMXFLOWWATCHERS 1, STREAMTIMEOUTS 0
   ua-trace:  DFMXFLOWWATCHERS 1, STREAMTIMEOUTS 1
   ua-live:   DFMXFLOWWATCHERS 4, STREAMTIMEOUTS 1

   Fri, 21 May 10 (NZST):  These are set by ./configure in owt_conf.h
      ./configure --enable-two-way --with-watchers=2 */

// #define DFMXFLOWWATCHERS   1  /* Number of flow watcher threads */
//                              /*   (4 has more overhead on odin) */
// #define STREAMTIMEOUTS     0  /* Time out inactive streams */

#if STREAMTIMEOUTS
#  define HALF_WAY_STREAMS   1  /* Watch 'half-way' streams;
      increasing TCP seq/ack => we're not seeing the return path */
#  define ERF_DIRECTION      1  /* 1 to count interface field from DAG card */
#else
#  define HALF_WAY_STREAMS   0
#  define ERF_DIRECTION      0
#endif

#define HALF_ONE_WAY_PKTS  10  /* If we see this many pkts with increasing
				  TCP seq or ack, it's a two-way stream,
				  but we're only seeing one direction of it */

#define MKSTATSCOUNTS      0  /* To build/collect interval stats */
#define MKDESTCOUNTS       1  /* To build/collect dst_addr/24block stats */
#define NDESTBINS        256  /* 256 /24 blocks in telescope space */

#define CHECKPKTQUEUE      0  /* Debugging defines */
#define STOP_AFTER         0  /* stop_meter() after this many #Stats calls */
                              /*   0 to only stop at EOF */
#define SC_DEBUG  0  /* Stream chain checks:
         0 = none
         1 = start of cto() [check_stream_timeouts()]
         2 = start cto() + print cto() stats
         3 = start & end of cto() + end of each_pkt() */
#define STARTUP_DEBUG  0  /* 1 to print 'got to here' lines */

#define STREAMS_TO_WRITE  0  /* 1 to write all stream summary info */


#define PRODUCTION         1  /* 0 to display statistics < every minute */
#if PRODUCTION
#define STATS_RECORDING_INTERVAL  60  /* 1m */
#else
#define STATS_RECORDING_INTERVAL  20  /* 20s */
#endif
#define STATS_INTERVALS_PER_HOUR  3600/STATS_RECORDING_INTERVAL
/* #Time recording interval is set by -c in owt.init method */

//#define METER_VERSION  "2.0"  /* Two-way timeouts, C source analysis */
#define METER_VERSION  "3.0"  /* Source IAT distributions, source groups */

#include "owt_queue.h"
                               /* Default sizes */
#define DFMXOPQENTRIES  25000  /* Size of output queue */
#define DFMXSTREAMS    250000  /* Nbr of streams */
#define DFMXEVENTS         10  /* Nbr of events */
#define DFMXSTREAMTXTS     20  /* Nbr of 'stream txt' messages */
#define DFMXPWQUEUESZ    1000  /* Max length of packet watcher input queue */
#define MXSTRMTXTLEN      501  /* Max length of a 'stream txt' message */
#define DFTRACESTOPROCESS   1  /* Number of trace files to process */

#define MXINTERFACES        2  /* Max interfaces we can watch */

#define MIN_HT_SIZE   2000000  /* Don't want long hash chains! */

#include "owt_mem.h"  /* Stream_engine memory management defines */

#define IP4_ADDR_LEN      4  /* IP address sizes */
#define IP6_ADDR_LEN     16

#define PT_ICMP        1  /* IP protocol types */
#define PT_TCP         6
#define PT_UDP        17
#define PT_SCTP      132
                          /* IPv6 protocol types */
#define PT_6HOP        0  /* Hop-by-hop options */
#define PT_6ROUT      43  /* Routing (Type 0) */
#define PT_6FRAG      44  /* Fragment */
#define PT_6ESP       50  /* Encapsulating Security Payload (ESP) RFC 2406 */
#define PT_6AH        51  /* Authentication (AH) RFC 2402 */
#define PT_6ICMP      58  /* ICMPv6 */
#define PT_6NONH      59  /* No next header */
#define PT_6DEST      60  /* Destination options */
#define PT_6MOBILE   135  /* IPv6 Mobility Header */ 


/* Well-known port numbers */
#define PN_SSH        22  /* ssh */
#define PN_TELNET     23  /* telnet */
#define PN_DNS        53  /* Domain Name System */
#define PN_US_DS     445  /* Microsoft-DS SMB file sharing */

#define ICMP_DST_UNREACHABLE   3  /* ICMP Types */
#define ICMP_TTL_EXCEEDED      11
#define ICMP6_DEST_UNREACHABLE  1  /* ICMP6 Types */
#define ICMP6_TTL_EXCEEDED      3

#define SKEY_OK           0  /* s_key return codes */
#define SKEY_NOSPACE      1
#define SKEY_NOTIP        2
#define SKEY_NOPEERADDR   3
#define SKEY_FRAGPKT      4
#define SKEY_NOTRANSADDR  5
#define SKEY_BADIPHDR     6
#define SKEY_V6SMALLSNAP  7
#define SKEY_V6OPAQUE     8
#define SKEY_NOLOCALADDR  9

extern char *logname;
extern FILE *log_file;
void log_msg(int priority, int die, char const *fmt, ...);


struct owt_ht {
   int table_size;  /* Max entries, used as hashmod */
   pthread_mutex_t ht_lock;  /* Lock for hash table */
   void **sfht;  /* Actual hash table */
   };

#define SS_FREE          0  /* Stream states */
#define SS_ONE_WAY       1
#define SS_TWO_WAY       2  /* Don't need to save pkts */
#define SS_HALF_WAY      3
#define SS_DELETE        4
#define SS_TIMEOUT       5
#define SS_n_states      6

#pragma pack(push,1)

struct v4_address {
   union {
      uint8_t a[IP4_ADDR_LEN];
      uint32_t v4;
      } a;
   };

struct v6_address {
   union {
      uint8_t a[IP6_ADDR_LEN];
      struct {
         uint64_t oct0, oct1;
         } o;
      } a;
   };

struct ip_address {
   union a {
      uint32_t v4;
      uint8_t a[IP6_ADDR_LEN];
      struct o {
         uint64_t oct0, oct1;
         } o;
      } a;
   uint8_t ver, len;  /* addr_is_local() needs ver */
   };

                          /* Parameters for source statistics */
#define MIN_LIFETIME 1.0  /* seconds */

extern int save_distributions;  /* Non-zero to print source distribs */
#define LOWER_IAT  0.15  /* s  IAT Distribution parameters */
#define UPPER_IAT  600.0
#define MXBUCKETS  120  // 180

struct distribution {
   struct distribution *next;

   double last_us;  /* For computing IATs */

   uint32_t total_count, bin0_count;
   uint16_t counts[MXBUCKETS+1];  /* +1 for oflo bucket */
   uint8_t Transform, Buckets;  /* Distrib parameters */
   double LowerLim, UpperLim;

   double M;   /* Range, for transform() */
   double step;  /* For computing bin edges */

#if 0
   uint8_t non_empty,  /* Sum of counts[] >= 1 */
      dyn_dist;  /* Using a dynamic distribution */
   uint32_t n_values,  /* Number of points in 'dynamic' distribution */
      lowerspec, upperspec,  /* Specified by user */
       min_val, max_val;  /* Observed in data */
#endif
   };

/* Transform values
       1     1    linear, counter -> histogram
       2     2    log,    counter -> histogram
       4     5    linear, dynamic -> set of values
       4     6    linear, dynamic -> histogram
       7     5    log,    dynamic -> set of values
       7     8    log,    dynamic -> histogram  */

/* 'Dynamic' distributions:
      Collects actual values until trigger point reached,
      then sets lower and upper parameters from data */

#  define DS_NULL       0  /* Distribution transform type values */
#  define DS_LIN        1  /* Linear */
#  define DS_LOG        2  /* Logarithmic */

#  define DS_DYN_REQ    4  /* Dynamic: request, linear */
#  define DS_DYN_PTS    5  /* Dynamic: values (Buckets = nbr of values) */
#  define DS_DYN_BKT    6  /* Dynamic: bucket counts (limits set from data) */
#  define DS_DLOG_REQ   7  /* Dynamic: request, log */
#  define DS_DLOG_BKT   8  /* Dynamic: bucket counts, log */

owt_queue_declare_globals(distribution);  /* Declare object space */
owt_mem_declare_globals(distribution);

#define MX_N_UINT32  10
#define MX_N_UINT16  10
#define MX_N_UINT8   10
#define V4_SET_SIZE   9  /* 19 Oct 11, allow for prev_dest */
#define V6_SET_SIZE   5

#define topn_t(name, sz, type) \
   struct topn_##name { \
      uint32_t nv; \
      uint32_t counts[sz];  type values[sz]; \
      };

topn_t(uint32, MX_N_UINT32, uint32_t);
topn_t(uint16, MX_N_UINT16, uint16_t);
topn_t(uint8, MX_N_UINT8, uint8_t);
topn_t(dst_addrs4, V4_SET_SIZE, struct v4_address);
topn_t(dst_addrs6, V6_SET_SIZE, struct v6_address);

struct v4_flow_info {
   struct v4_flow_info *next;
   struct distribution *d;
   uint32_t cfc_count, uT_count, tcp_key;
   struct topn_uint16 tn_tcp, tn_udp, tn_src_port;
   struct topn_uint8 tn_flags, tn_proto, tn_ttl;
   struct topn_dst_addrs4 tn_dst_addrs;
#if HALF_WAY_STREAMS
   struct v4_address prev_dst;
   uint32_t pkts_this_dst, prev_seq, prev_ack;
#endif
   };
owt_queue_declare_globals(v4_flow_info);
owt_mem_declare_globals(v4_flow_info);

struct v6_flow_info {
   struct v6_flow_info *next; 
   struct distribution *d;
   uint32_t cfc_count, uT_count, tcp_key;
   struct topn_uint16 tn_tcp, tn_udp, tn_src_port;
   struct topn_uint8 tn_flags, tn_proto, tn_ttl;
   struct topn_dst_addrs6 tn_dst_addrs;
#if HALF_WAY_STREAMS
   struct v6_address prev_dst;
   uint32_t pkts_this_dst, prev_seq, prev_ack;
#endif
   };
owt_queue_declare_globals(v6_flow_info);
owt_mem_declare_globals(v6_flow_info);

extern uint16_t probe_port;  /* Results from analyse_flow() */
extern int flags_mode, keys_mode;
extern uint8_t probe_port_valid;   /* (other than src_type) */
extern uint8_t too_few_pkts;     /* < MinPkts */
extern uint8_t protocol;

//#define PD_FREE          0   /* free_stream */
                             /* 1-5 are the SS states */
//#define PD_READ    SS_n_states   /* by get_v*key */
//#define PD_n_states  PD_READ+1

struct stream {  /* Hash table entry, key is (src/dest addr) */
   struct stream *next_hc;    /* For utht(), must be first */
   uint8_t version, state,    /* For utht(), should be second */   
      first_fwd,  /* Remembers fwd value for first packet */
      timeout;    /* In seconds for this stream */

   struct stream *next, *prev;  /* For stream queue */
   struct ip_address src_addr;
   uint32_t last_report_sec;  /* ts when last stream report was queued */

  /* Summary data for the source ... */
   double first_us, last_us;  /* Times for first and last packets */
   uint64_t tb, fb;           /* To/From byte counts */
   uint32_t tp, fp, totp;      /* To/From packet counts */
   union {
      struct v4_flow_info *fi4;
      struct v6_flow_info *fi6;
      } flow_info;

   /* Data for individual packet (or event) ... */
      /* last_us = pkt arrival time in us */
   struct ip_address dst_addr;  /* For 'this packet' */
   uint16_t src_port, dst_port, ip_len;
   uint8_t proto, tcp_flags, cfc, uTorrent,
#if ERF_DIRECTION
     ttl, res1, res2, direction;
#else
     ttl, res1, res2, res3;
#endif
   uint32_t tcp_key;  /* mss, window */
#if HALF_WAY_STREAMS
   uint32_t tcp_seq, tcp_ack;
#endif
   uint16_t wl;         /* Packet's wire length */
   uint32_t ts;         /* Seconds field of packet's timestamp */
   uint32_t hash;       /* 32-bit hash for sp */
   uint8_t event_type,  /* pkt event (_EVENT values, 0 to just count pkt) */
      fwd;              /* fwd value for this packet */

   uint8_t  /* Variables used by the Ruby outer block ... */
      fw_nbr,          /* flow_watcher nbr */
      ext_blk,  /* 0 = stream_space, >0 = extra_streams[] */
      active_since_last_sum,
      inactive_last_sum;
#if MKSTATSCOUNTS 
   uint16_t last_active_interval;
#endif
   };

extern uint16_t probe_port, src_port;  /* analyse_flow() results */
extern int flags_mode, keys_mode;
extern uint8_t probe_port_valid,   /* (other than src_type) */
   too_few_pkts,      /* < MinPkts */
   protocol;

extern double b0pc, modeiat, skew;  /* compute_distrib_stats() results */
extern int nz_runs;

void init_distrib(struct distribution *d);
void bump_dist(  /* Add point y into distrib d */
   struct distribution *d, double y);
void accum_dist(  /* Add v to point y of distrib d */
   struct distribution *d, double y, uint32_t v);
// void end_burst(struct distribution *d);
void write_distrib_info(FILE *f, struct stream *sp, struct distribution *d);
int compute_distrib_stats(struct stream *sp);
void dump_source_info(FILE *f, struct stream *sp, int src_type,
   struct distribution *d);

owt_queue2_declare_globals(stream);
owt_mem_declare_globals(stream);

#define MX_EXTRA_BLOCKS     2  /* stream, flow_info and distribution */
extern struct stream *extra_streams[MX_EXTRA_BLOCKS+1];
extern int extra_stream_incr, n_extra_streams;
#define MN_FREE_STREAMS     200000
#define MN_STREAMS_TO_FREE  250000

#define TOO_MANY_TIMEOUT     120  /* 'Flood' stream time to be "idle" */
#define MN_PKTS_FOR_DISCARD    2  /* recover streams with more pkts */

extern uint32_t n_unanalysed_sources, n_unanalysed_pkts;
extern uint32_t    /* Stats for unanalysed sources */
   ua_pkts[MN_PKTS_FOR_DISCARD],  /* Nbr of packets */
   ua_src_addr[256],  /* Source addresses */
   ua_dst_addr[256],  /* Dest addresses */
   ua_ttl[256],       /* TTLs */
   ua_proto[256],     /* Protocols */
   ua_gt1prot;        /* More than one proto seen */

#pragma pack(pop)

#if MKSTATSCOUNTS
extern uint32_t  /* Per-interval stats arrays for each source type .. */
//x   **pkts_seen_per_stats_interval, *pkts_seen_this_interval,
   **active_sources_per_stats_interval, *active_sources_this_interval,
   **unique_hosts_per_stats_interval, *unique_hosts_this_interval;
extern uint32_t stats_interval_nbr;
#endif
#if MKDESTCOUNTS
extern uint32_t **dest_addrs_per_source_type;
#define src_addr_bin(a)  ((ntohl(a) >> 24) & 0x000000FF)  /* 1st byte */
extern uint32_t (*dest_addr_bin)(uint32_t a);
#endif

#define MX_LOCAL_ADDRS  5
extern struct ip_address la[MX_LOCAL_ADDRS];
extern int n_local_addrs;

#define TCP_STREAM_TIMEOUT     2  /* 2s (FIN_WAIT) */
#define STO_MULTIPLIER        10

#define MX_ONE_WAY_TIMEOUTS_PER_CHECK  20000  /* 6 Aug 10 */
   /* Limit demand on opq_empty!  23 Jun 10 */

#define TWO_WAY_TIMEOUT        5  /*  5s */
#define ONE_WAY_TIMEOUT       10  /* 10s */

#define STO_TWO_WAY           2  /* Timeout kinds for streams .. */
#define STO_ONE_WAY           1

extern uint32_t n_source_types, pkts_to_record, min_pkts_to_classify,
   traces_to_process;
extern char **other_uris;
extern int next_uri_nbr;


extern int time_rec_interval, intervals_this_sum_file;

extern int n_lost_stream_v4s, n_lost_stream_v6s;
extern uint64_t last_drops;

extern uint32_t mx_active_opq_entries, mx_active_streams,
   mx_active_v4_flow_infos, mx_active_v6_flow_infos,
   av_cl_count, n_hc_searches,  src_av_cl_count, src_n_hc_searches;
extern uint32_t n_not_ip,n_frags,n_too_small,n_no_local_address,
   n_v6_opaque;

extern uint64_t  /* Overall stream summary info */
   v4_one_way_bytes,  v4_one_way_pkts,  v4_one_way_streams,
   v6_one_way_bytes,  v6_one_way_pkts,  v6_one_way_streams,
   v4_two_way_bytes,  v4_two_way_pkts,  v4_two_way_streams,
   v6_two_way_bytes,  v6_two_way_pkts,  v6_two_way_streams,
   v4_half_way_bytes, v4_half_way_pkts, v4_half_way_streams,
   v6_half_way_bytes, v6_half_way_pkts, v6_half_way_streams;
#if ERF_DIRECTION
extern uint64_t
   one_way_streams_1, one_way_streams_2, one_way_streams_3,
   two_way_streams_1, two_way_streams_2, two_way_streams_3, 
   half_way_streams_1, half_way_streams_2, half_way_streams_3;
#endif

extern uint32_t start_ts,  /* Seconds part of first-packet */
   last_ts;                /*    and last-packet timestamps */

struct if_info {
   uint32_t n_pkts,  /* Packets this second */
      n_seconds,     /* Seconds since start of interval */
      mx_pkt_rate, av_pkt_count;  /* Max/s, total pkts this interval */
   uint32_t last_ts;
   char *if_name;
   int thread_running;
   pthread_t pw_thread;  /* Thread for input packet watcher */
   libtrace_t *trace;
   libtrace_packet_t *packet;
   int if_nbr;
   };

extern int snaplen;
extern uint32_t g_last_ts;  /* Max last_ts for all interfaces */

extern struct if_info if_info[MXINTERFACES];
extern uint32_t n_interfaces,   /* Nbr of interfaces specified */
   n_interfaces_running;
extern char *trace_name;  /* URI for interface 0 */
extern char *meter_loc;   /* SAN, AKL, ... */

#define OQ_NULL           0  /* OQ (Output Queue) entry types */
#define OQ_TXTMSG         1
#define OQ_STARTED        2  /* Not used 21 Aug 10 */
#define OQ_PROGRESS       3  /* Not used 21 Aug 10 */
#define OQ_ENOUGH         4
#define OQ_TIMED_OUT      5
#define OQ_END_OF_FILE    6

struct opq_entry {
   uint8_t type;  /* OQ_ type */
   int s_id;      /* Index in stream or src_host array */
   char *msg;     /* For 'stream txt' message */
   time_t ts;     /* ts seconds for txt msg */
   struct opq_entry *next; 
   };

owt_queue_declare_globals(opq_entry);
owt_mem_declare_globals(opq_entry);

extern owt_queue_declare_space(opq, opq_entry);  /* owt output queue */

#define NO_EVENT       0  /* Event types */
#define S_TIME_EVENT   1  /* Check stream timeouts */
#define STATS_EVENT    2  /* Write interface stats */
#define TIME_EVENT     3  /* Write #Time record */
#define END_OF_FILE    4  /* End of trace file */

struct event {
   uint32_t t;
   uint16_t type, interval;  /* Interval in seconds */
   struct event *next;
   };
extern struct event event_q;  /* Queue of active events, dummy first member */

owt_queue_declare_globals(event);
owt_mem_declare_globals(event);

struct stream_txt {
   char msg[MXSTRMTXTLEN+1];
   struct stream_txt *next;
   };

owt_queue_declare_globals(stream_txt);
owt_mem_declare_globals(stream_txt);

/* mutexes for data structures shared between threads .. */
extern pthread_mutex_t 
   log_file_mutex,     /* log_file */
   event_q_mutex,      /* Event queue */
   if_info_mutex,      /* if_info table */
   stream_data_mutex,  /* stream data (packet/byte counts) */
   n_fw_mutex,         /* n_fw_running */
   count_mutex;        /* lock for count/count_cv */

extern  pthread_cond_t count_cv;

extern int count;  /* Nbr of check_timeouts() running */
#define CV_CHECKING   (count > 0)
#define CV_IDLE       (count == 0)
#define CV_RECORDING  (count < 0)

struct flow_watcher_control {
   struct flow_watcher_control *next;
   owt_queue_declare_space(ispq, stream);  /* Packet-watcher output queue */
   Semaphore ispq_ready, ispq_empty;  /* Flow watcher input queue */
  //n   struct owt_ht *src_table;  /* Src_host (hash) table */
   struct owt_ht *sh_table;  /* Stream (hash) table */
   owt_queue2_declare_space(sq, stream);  /* (Bi-directional) stream queue */
   struct stream *eof_sp;  /* So packet-watcher can signal EOF */
   pthread_t flow_thread;
   int id;
   };

owt_queue_declare_globals(flow_watcher_control);
owt_mem_declare_globals(flow_watcher_control);

extern int n_fw_running;  /* Nbr of flow watchers running */

extern Semaphore stream_empty;  /* Controls for packet-watcher input queue */

extern Semaphore opq_ready, opq_empty;  /* Output queue (to outer-block loop) */

#include <sys/utsname.h>
extern struct utsname node;  /* Info about the host we're nunning on */

extern char *version;

/* Functions in owt_watcher.c */

void quack(int nnn);
int sv4p_bad(struct stream *sp);

char *owt_file_header(void);

void update_interface_stats(uint32_t not_ts);
void each_time(uint32_t now_ts);
void check_stream_timeouts(struct flow_watcher_control *fw, uint32_t now_ts);
void output_active_streams(struct stream_queue *sq, uint32_t now_ts);

void one_second_process(uint32_t ts);
char *fmt_time(int utc_time, uint32_t *ti);
extern int request_shutdown;  /* True to stop stream engine */
void start_meter(char *interfaces[], uint32_t n_ifs,
   uint32_t mx_events, uint32_t mx_opq_entries, uint32_t mx_streams,
   uint32_t mx_pw_q_sz, uint32_t mx_stream_txts, uint32_t mx_flow_watchers,
   uint32_t mx_source_types, uint32_t hours_reqd);
void stop_meter(char const *msg);

void each_pkt(struct flow_watcher_control *fw, struct stream *isp);
void cancel_packet_watchers(void);

#include <unistd.h>

#define SNAP_LEN  120

struct ether_hdr {  
   uint8_t dest[6];    /* MAC addresses */
   uint8_t source[6];
   uint8_t type[2];    /* Length or protocol type */
   };

extern int live_source;  /* Zero for trace files */

double at_seconds(struct timeval tv);
uint32_t pkt_drops(void);
int init_trace(struct if_info *iip);
void close_trace(struct if_info *iip);
void *packet_watcher_process(void *arg);
void *flow_watcher_process(void *arg);

extern int errno;
void usage(char *argv0);

#include <getopt.h>
int main(int argc, char *argv[]);

/* Functions in owt_monitor.c */

/* Functions in owt_support.c */

#include <pwd.h>
void uid_from_cwd(void);
extern int my_uid, ut_uid, my_gid, ut_gid;

#include <stdarg.h>

struct event *get_event(int t, int interval);
void schedule_event(struct event *ev, uint32_t first_t);
int check_events(uint32_t now);

struct opq_entry *oqe_for_stream(struct stream *sp);
struct stream *stream_from_oqe(struct opq_entry *oqe);

void get_flow_info(struct stream *sp);
void init_flow_info(struct stream *sp);
void free_flow_info(struct stream *sp);

struct stream *get_stream(int kind);
void init_get_info(struct stream *sp);
void init_flow_info(struct stream *sp);
void free_flow_info(struct stream *sp);
void free_stream(struct stream *sp);
void stream_fl_check(void);

void init_mutex(pthread_mutex_t *m, char const *name);
void _mutexes(void);

#define mutex_lock(caller,m)  tw_pthread_mutex_lock(m)
#define mutex_unlock(caller,m)  tw_pthread_mutex_unlock(m)
#define ntm_semaphore_down(caller,s)  t_semaphore_down(s)
#define ntm_semaphore_up(caller,s)  t_semaphore_up(s)
#define ntm_semaphore_init(name,id,s)  t_semaphore_init(s)

void owt_start(int mx_events, int mx_opq_entries, int mx_streams,
   int mx_stream_txts, int mx_flow_watchers);
void owt_shutdown(void);

#define v6cmp(s, d)  memcmp(s.a, d.a, IP6_ADDR_LEN)
char *v6addr_to_s(char *v6a, struct ip_address *addrp);

struct owt_ht *owt_ht_create(int mxhtentries);
uint32_t owt_hash(struct stream *record);
void *owt_ht_lookup(struct owt_ht *t, uint32_t hash,
   struct stream *record, int *OK_to_add);
   /* returns record, NULL if not found */
void owt_ht_delete(struct owt_ht *t, void *record);
   /* deletes specified record */

int addr_is_local(int ver, struct ip_address *addrp);

void reset_summary_info(void);
void update_summary_info(struct stream *sp);

void count_uint16(uint16_t p, struct topn_uint16 *tnp);
void count_uint8(uint8_t p, struct topn_uint8 *tnp);
void count_addrs4(struct ip_address *p, struct topn_dst_addrs4 *tnp);
void count_addrs6(struct ip_address *p, struct topn_dst_addrs6 *tnp);
int update_flow_info(struct stream *sp, struct stream *isp);

uint16_t top_uint16(struct topn_uint16 *tnp);
uint8_t top_uint8(struct topn_uint8 *tnp);
uint32_t top_addrs4(struct topn_dst_addrs4 *tnp);
int analyse_flow(struct stream *sp);

extern char key_s[150];  /* Return value for key_to_s() et al */
char *key_to_s(struct stream *sp);
void print_topn_dst_addrs4(struct topn_dst_addrs4 *tnp);
void print_topn_dst_addrs6(struct topn_dst_addrs6 *tnp);
void print_topn_uint8(struct topn_uint8 *tnp);
void print_topn_uint16(struct topn_uint16 *tnp);
void print_flow_info(struct stream *sp);

void update_ua_stats(struct stream *sp);


#include "hashtable.h"  /* C hashtable by Christopher Clark
                           https://github.com/davidar/c-hashtable */
#include "hashtable_itr.h"

#if HALF_WAY_STREAMS  /* For src_addr ht */
struct addr_key { struct ip_address key; };
struct addr_key_val { struct ip_address key; uint32_t count; };
extern struct hashtable *half_src_addrs;

static unsigned int hashfromaddr(void *ky);
static int equaladdrs(void *ky1, void *ky2);
static void addr_ht_insert(struct hashtable *h, struct ip_address *addr);

static int addr_kv_cmp(const void *a, const void *b);
struct addr_key_val *addr_get_kv( struct hashtable *h, int *length);
#endif

#if ERF_DIRECTION
void update_direction_counts(struct stream *sp);
#endif

struct key { uint32_t key; };  /* For port and tcp_key hts */
struct key_val { uint32_t key, count; };
extern struct hashtable *ua_dst_ports, *ua_tcp_keys;

static unsigned int hashfromkey(void *ky);
static int equalkeys(void *ky1, void *ky2);
static void owt_ht_insert(struct hashtable *h, uint32_t tk);
static int kv_cmp(const void *a, const void *b);
struct key_val *get_kv( struct hashtable *h, int *length);
