#!/usr/bin/env ruby

#  2015, Tue  1 Jun 10 (NZST)
#  1622, Tue 13 Apr 10 (PDT)
#  1425, Thu 31 Dec 09 (NZDT)
#
#  newmon.rb: OWT monitor, with C source classifier
#
#  Copyright 2007-2011, Nevil Brownlee, U Auckland | CAIDA | WAND
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "#{File.dirname($0)}/owt-config"

begin
  require "./IPprefix"  # Is it installed so Ruby can find it?
rescue LoadError
  require "#{Home}/IPprefix"  # No, try for it in Home directory
end

require "#{Home}/owt-c/OWTmonitor"  # Must follow IPprefix (!)
require "#{Home}/owt-rb/owt-support" 
require "#{Home}/owt-rb/owt-monitor" 
require "#{Home}/owt-rb/distrib"

# ./owtmon.rb dag:/dev/dag0 sum2                                  # ua-live
# ./owtmon.rb traces/ua.20091201.1700.bpf.gz sum2 sum2/san-trace  # ua-trace
# ./owtmon.rb /work/telescope/live/2009-12-21/net44.1261404000.pcap.gz sum2 sum2/test/2009-12-21.fdf  # odin

# ./newmon.rb pcapfile:/home/nevil/ua64-traces/test20101213.011030 sum2
# ./newmon.rb pcapfile:/scratch/owtmon/net44.1299596400.pcap.gz sum2 test.fdf 3 D

trace_URI = ARGV[0]    # URI of trace to process
output_dn = ARGV[1]    # output sum2 directory 
log_fn = ARGV[2]       # name of owtmon log file (appended to)

$clog_level = ARGV.size < 4 ? 3 : ARGV[3].to_i
  # 0 or empty = display all log messages
  # 1          = display Time messages
  # 2          = display Time and Stats messages

distribs_reqd = "-u#{ARGV[4]}" if !ARGV[4].nil?
  # D to write distributions, one for each one-way source
  # T to only write types+groups file (no sum2 file)

trace_URI = SourceURI  if trace_URI.nil?
output_dn = SummaryDir if output_dn.nil?
log_fn = StatsFile     if log_fn.nil?

OWTmonVer = '3.1';  FormatVer = '3.1'

TCP = 6;  UDP = 17;  ICMP = 1;  ICMP6 = 58  # Protocol numbers
ICMP_DEST_UNREACH  =   3  # ICMP Types
ICMP_TTL_EXCEEDED  =  11
ICMP6_DEST_UNREACH =   1  # ICMP6 Types
ICMP6_TTL_EXCEEDED =   3
# TCP flags  32 URG, 16 ACK, 8 PSH, 4 RST, 2 SYN, 1 FIN 

def file_dn(output_dn, ts)
  check_dir_exists(output_dn)
  odn = "#{output_dn}/#{utc_hts(ts)}"
  check_dir_exists(odn)
  odn
end

def day_change(ts)
  if $today_date.nil?
    $today_date = utc_hts(ts)
    return true
  else
    new_date = utc_hts(ts)
    change = new_date != $today_date
    $today_date = new_date
    return change
  end
end

def next_uri(uri)  # Get filename for next telescope trace file
  uri =~ /^(.+):(.+)$/
  utype = $1;  fn = $2
  arg_fn = File.basename(fn)
  
  fn =~ /^(.+)\/(20.+)\/(.+)\.(\d+)\.(.+)$/
  live_dir = $1;  sdate = $2;  net = $3;  ut = $4;  ftype = $5

  return_next = false
  Dir.glob("#{live_dir}/????-??-??").sort.each do |an|
    date = File.basename(an) 
    next if date < sdate
    Dir.glob("#{live_dir}/#{date}/#{net}.*.#{ftype}").sort.each do |fulln| 
      fn = File.basename(fulln)
      return "#{utype}:#{fulln}" if return_next
      return_next = true if fn == arg_fn
    end
  end
end

class NbrHash < Hash  # For port numbers and ip lengths
  def update(str)
    if !self[str]
      self[str] = 1
      return true  # new
    else
      self[str] += 1
      return false
    end
  end

  def val_range
    min = 65535;  max = 1
    self.each do |key, val|
      ikey = key.to_i
      min = ikey if ikey < min
      max = ikey if ikey > max
    end
    [self.size, min, max]
  end

  def mode
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    shash[0][0].to_i
  end

  def write(f)
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    n = 0
    shash.each do |kv|
      f.printf("  %u %u", kv[0], kv[1])
      n += 1
      break if n == N_ProbePorts
    end
  end

  def write_str(f, n_reqd)
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    n = 0
    shash.each do |kv|
      f.print"  #{kv[0]} #{kv[1]}"
      n += 1
      break if n == n_reqd
    end
  end
end  # NbrHash

N_source_types = SrcTypeNames.size

def write_sum_file(fname, start_ts, now_ts)
  $osf = File.new(fname, 'w')
  $osf.print "#version\n"
  $osf.print "#{FormatVer} #{OWTmonVer} #{Meter_ver}\n\n"

  $osf.print "#meter_info\n"
  $osf.print "#{MeterLoc} #{utc_gts(start_ts)}" +
    " #{Trace_sz} #{MinPkts} #{TwoWay ? 'two-way' : ''}\n\n"

  src_counts = Array.new(SrcTypeNames.size)  # rows S_ types
  byte_counts = Array.new(SrcTypeNames.size)
  src_counts.each_index do |r|  # cols IAT_ groups
    src_counts[r] = Array.new(GroupNames.size, 0)
    byte_counts[r] = Array.new(GroupNames.size, 0.0)
  end

  src_type_counts_v4 = Array.new(N_source_types, 0)
  tf_type_counts_v4 = Array.new(N_source_types, 0)
  if LocalV6
    src_type_counts_v6 = Array.new(N_source_types, 0)
    tf_type_counts_v6 = Array.new(N_source_types, 0)
  end

  tcp_probe_port_counts = NbrHash.new;  udp_probe_port_counts = NbrHash.new
  proto_counts = NbrHash.new

  lifetime_distribs = Array.new;  pkt_rate_distribs = Array.new
  tot_count_distribs = Array.new
  pkt_count_distribs = Array.new;  tf4_pkt_count_distribs = Array.new
  if LocalV6
    v6_top32_counts = NbrHash.new
    v6_pkt_count_distribs = Array.new;  tf6_pkt_count_distribs = Array.new
  end

  (0..N_source_types-1).each do
    lifetime_distribs.push(Distribution.new(1.0, MxLifetime, 120, DT_LOG, 1.0))
    pkt_rate_distribs.push(Distribution.new(0.001, 1000.0, 100, DT_LOG, 1.0))
    pkt_count_distribs.push(Distribution.new(1, 5000, 100, DT_LOG, 1.0))
    if TwoWay
      tot_count_distribs.push(Distribution.new(1, 50000, 120, DT_LOG, 1.0))
    end
    tf4_pkt_count_distribs.push(Distribution.new(1, 10, 10, DT_LINEAR, 1.0))
    if LocalV6
      v6_pkt_count_distribs.push(Distribution.new(1, 5000, 100, DT_LOG, 1.0))
      tf6_pkt_count_distribs.push(Distribution.new(1, 10, 10, DT_LINEAR, 1.0))
    end
  end

  msg = $owt.each_stream do |s|  # look at final attribs for all streams
    src_type, probe_port, to_pkts, total_lifetime, tot_count, version,
      too_few_pkts, protocol, src_group, flags_mode, keys_mode,
      v6_addrs_s32, to_bytes  = s

    src_counts[src_type][src_group] += 1
    byte_counts[src_type][src_group] += to_bytes/1000000.0

    if version == 4
      if too_few_pkts
        tf_type_counts_v4[src_type] += 1
        tf4_pkt_count_distribs[src_type].count(to_pkts)
      else
        src_type_counts_v4[src_type] += 1
        pkt_count_distribs[src_type].count(to_pkts)
      end
    elsif LocalV6
      v6_top32_counts.update(sprintf("%08x", v6_addrs_s32))
      if too_few_pkts
        tf_type_counts_v6[src_type] += 1
        tf6_pkt_count_distribs[src_type].count(to_pkts)
      else
        src_type_counts_v6[src_type] += 1
        v6_pkt_count_distribs[src_type].count(to_pkts)
      end
    end
    if TwoWay
      tot_count_distribs[src_type].count(tot_count)
    end
    if !probe_port.nil?  # && !too_few_pkts
        # 'probe' means 'all pkts sent to same port'
        # even if source sends only one pkt!  28 Apr 11
      if src_type == S_tcp_probe
        tcp_probe_port_counts.update(probe_port)
      elsif src_type == S_udp_probe
        udp_probe_port_counts.update(probe_port)
     end
    end
    proto_counts.update(protocol)
    lifetime_distribs[src_type].count(total_lifetime)
    total_rate = to_pkts < MinPkts ? 0 : (tot_count - 1)/total_lifetime
    pkt_rate_distribs[src_type].count(total_rate)
  end
  way_counts, st_distribs, half_src_addrs, half_src_counts =
     $owt.c_addr_distribs  # Must call this *after* .each_stream{}!

  if !msg.nil?  # #Streams
    $owt.log_msg(0, msg)  # Write to log
    $stdout.flush if $clog_level > 0  # and console
  end

  $osf.print "#MB_counts\n"
  byte_counts.each_index do |r|  # rows (types)
    byte_counts[r].each_index do |c|  # cols (groups)
      $osf.printf("%12.6f ", byte_counts[r][c])
    end
    $osf.print "\n"
  end
  $osf.print "\n"

=begin
ra = $owt.unanalysed
ra.each_index do |x|
  print "ra[#{x}] = #{ra[x]}, class = #{ra[x].class}\n"
end
=end
  ua_sources, ua_totpkts, ua_pkts, ua_src_addr, ua_dst_addr, ua_ttl, ua_proto, ua_gt1prot,
    ua_dst_port_keys,ua_dst_port_counts, ua_tcpkey_keys,ua_tcpkey_counts  = $owt.unanalysed

  $osf.print "#source_counts  #{ua_sources} sources, #{ua_totpkts} pkts not analysed\n"
  src_counts.each_index do |r|  # rows (types)
    src_counts[r].each_index do |c|  # cols (groups)
      $osf.printf("%12d ", src_counts[r][c])
    end
    $osf.print "\n"
  end
  $osf.print "\n" 

  if ua_sources != 0
    $osf.print "#unanalysed\n"
    $osf.print "pkts ";  ua_pkts.each { |c| $osf.print " #{c}" };  $osf.print "\n"
    $osf.print "src_addr ";  ua_src_addr.each { |c| $osf.print " #{c}" };  $osf.print "\n"
    $osf.print "dst_addr ";  ua_dst_addr.each { |c| $osf.print " #{c}" };  $osf.print "\n"
    $osf.print "ttl ";  ua_ttl.each { |c| $osf.print " #{c}" };  $osf.print "\n"
    $osf.print "proto ";  ua_proto.each { |c| $osf.print " #{c}" };  $osf.print "\n"
    $osf.print "gt1prot #{ua_gt1prot}\n"
    $osf.print "dst_port  #{ua_dst_port_keys.size}"
    ua_dst_port_keys.each_index do |x|
      break if x == N_UnanalysedKeys
      $osf.printf("  %u %u", ua_dst_port_keys[x], ua_dst_port_counts[x])
    end
    $osf.print "\n"
    $osf.print "tcp_key  #{ua_tcpkey_keys.size}"
    ua_tcpkey_keys.each_index do |x|
      break if x == N_UnanalysedKeys
      $osf.printf("  %08x %u", ua_tcpkey_keys[x], ua_tcpkey_counts[x])
    end
    $osf.print "\n\n"
  end

  $osf.print "#source_types\n"
  $osf.print N_source_types
  src_type_counts_v4.each { |c| $osf.print " #{c}" }
  $osf.print "\ntf4"
  tf_type_counts_v4.each { |c| $osf.print " #{c}" }
  if LocalV6
    $osf.print "\nv6"
    src_type_counts_v6.each { |c| $osf.print " #{c}" }
    $osf.print "\ntf6"
    tf_type_counts_v6.each { |c| $osf.print " #{c}" }
  end
  $osf.print "\n\n"

  $osf.print "#tprobe_ports\n#{N_ProbePorts}"
  tcp_probe_port_counts.write($osf)
  $osf.print "\n\n"

  $osf.print "#uprobe_ports\n#{N_ProbePorts}"
  udp_probe_port_counts.write($osf)
  $osf.print "\n\n"

  $osf.print "#protocols\n#{proto_counts.size}"
  proto_counts.write($osf)
  $osf.print "\n\n"

  if LocalV6
    n_v6_slash32s = 20
    $osf.print "#v6addrs\n#{n_v6_slash32s}"
    v6_top32_counts.write_str($osf, n_v6_slash32s)
    $osf.print "\n\n"
  end

  $osf.print "#lifetimes\ns "
  lifetime_distribs[0].write_x($osf, 4)
  lifetime_distribs.each_index do |st|
    $osf.print "#{st} "
    lifetime_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  $osf.print "#pktrates\np/s "
  pkt_rate_distribs[0].write_x($osf, 4)
  pkt_rate_distribs.each_index do |st|
    $osf.print "#{st} "
    pkt_rate_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  $osf.print "#pktcounts (this hour)\ni "
  pkt_count_distribs[0].write_x($osf, 1)
  pkt_count_distribs.each_index do |st|
    $osf.print "#{st} "
    pkt_count_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  if LocalV6
    $osf.print "#v6pktcounts\ni "
    v6_pkt_count_distribs[0].write_x($osf, 1)
    v6_pkt_count_distribs.each_index do |st|
      $osf.print "#{st} "
      v6_pkt_count_distribs[st].write_y($osf, false)
    end
    $osf.print "\n"
  end

  $osf.print "#tf4pktcounts\ni "
  tf4_pkt_count_distribs[0].write_x($osf, 1)
  tf4_pkt_count_distribs.each_index do |st|
    $osf.print "#{st} "
    tf4_pkt_count_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  if LocalV6
    $osf.print "#tf6pktcounts\ni "
    tf6_pkt_count_distribs[0].write_x($osf, 1)
    tf6_pkt_count_distribs.each_index do |st|
      $osf.print "#{st} "
      tf6_pkt_count_distribs[st].write_y($osf, false)
    end
    $osf.print "\n"
  end

  if TwoWay
    $osf.print "\n"
    $osf.print "#totcounts (source lifetime)\nB  "
    tot_count_distribs[0].write_x($osf, 1)
    tot_count_distribs.each_index do |st|
      $osf.print "#{st} "
      tot_count_distribs[st].write_y($osf, false)
    end
    $osf.print "\n"

    $osf.print "#waycounts\n"
    $osf.print "#{way_counts[0]}, #{way_counts[1]}"    # 1-way pkts, bytes
    $osf.print "  #{way_counts[2]}, #{way_counts[3]}"  # 2-way pkts, bytes
    $osf.print "  #{way_counts[4]}, #{way_counts[5]}\n"  # 1 and 2-way streams
    $osf.print "#{way_counts[6]}, #{way_counts[7]}"    # v6 1-way pkts, bytes
    $osf.print "  #{way_counts[8]}, #{way_counts[9]}"  # v6 2-way pkts, bytes
    $osf.print "  #{way_counts[10]}, #{way_counts[11]}\n"  # v6 1 & 2-way streams
    if !way_counts[12].nil?
      $osf.print "\n#halfwaycounts\n"
      $osf.print "#{way_counts[13]}, #{way_counts[12]}, #{way_counts[14]}"
        # v4 half-way pkts, bytes, streams
      $osf.print ",  #{way_counts[16]}, #{way_counts[15]}, #{way_counts[17]}\n"
        # v6 half-way pkts, bytes, streams
      ntp = 16
      $osf.print "#{half_src_addrs.size}: "
      (0..(ntp-1)).each do |x|
        $osf.print ' ' if x != 0
        $osf.print "#{half_src_addrs[x]} (#{half_src_counts[x]})"
      end
      $osf.print "\n"
    end
    $osf.print "\n"
  end

fin_proc_time = Process::times;  fin_proc_date = Time.now
  time_string = sprintf("(minutes): trace_duration=%.2f" +
      ", elapsed=%.3f, user=%.3f, system=%.3f",
    (now_ts - start_ts)/60.0,
    (fin_proc_date.to_i - $start_proc_date.to_i)/60.0,
    (fin_proc_time.utime - $start_proc_time.utime)/60.0,
    (fin_proc_time.stime - $start_proc_time.stime)/60.0 )
  $start_proc_time = Process::times;  $start_proc_date = Time.now
  $osf.print "#times #{time_string}\n"
  if $clog_level > 0
    print "#Summary: times #{time_string}\n";  $stdout.flush
  end

  $osf.print "\n#eof \n"
  $osf.close
end  # write_sum_file()


#class OWTevent
#  attr_reader :id
#end

class OWTmessage
  attr_reader :type, :source, :text
end

# sudo dagfour -d /dev/dag0 -aei
# sudo dagfour slen=80 varlen
# sudo dagfour -d /dev/dag0 disableb
# sudo dagfour -d /dev/dag0 enablea  # A interface sees UA border traffic
# sudo dagsnap -v -o tracefile  # Check it's working

Type_s = {OQ_TXTMSG => 'OQ_TXTMSG', 
  OQ_TIMED_OUT => 'OQ_TIMED_OUT',
  OQ_END_OF_FILE => 'OQ_END_OF_FILE' }

$stderr.print \
  "trace_URI=#{trace_URI}, output_dn=#{output_dn}, log_fn=#{log_fn}\n"

$owt = Monitor.new("-i#{trace_URI} -s#{N_Streams} -n#{MinPkts} -t#{N_source_types} -c#{Hours_reqd*3600} -h#{Hours_reqd} -l#{MeterLoc} -o#{log_fn} #{distribs_reqd}")
addr_list = Array.new
v6_in_list = false
LocalAddress.split(/,\s*/).each do |la|

  as = IPprefix.from_s(la)
  print "as = #{as}, ver=#{as.version}, len=>#{as.length}\n";  $stdout.flush

  addr_list.push(IPprefix.from_s(la))
  v6_in_list = true if la =~ /:/
end

LocalV6 = v6_in_list
$owt.set_local_addrs( addr_list )

(Meter_ver, TwoWay) = $owt.info
print "Meter_ver=#{Meter_ver}, TwoWay=#{TwoWay}\n"

LiveTrace = $owt.live
if LiveTrace
  Trace_sz = Host  # Dummy for trace from live interface
else
  if $owt.n_interfaces != 1
    $stderr.print "Can only process one file per owtmon.rb run!\n"
    exit
  else
    Trace_sz = sprintf("%.3f", File.size($owt.last_fn)/1000000000.0)  # GB
  end
end

def sigint
  $stderr.print "^C pressed: stopping owtmon ...\n"
  $owt.close
end

#trap(:INT, lambda {sigint})  # ^C (Interrupt) =>  Set request_shutdown [1.8]
trap(:INT) { sigint }  # ^C (Interrupt) =>  Set request_shutdown [1.9]

#def sigtstp
#  $stderr.print "SIGTSTP (^Z pressed) ..\n"
#  st = caller(3)  # Ignore first 3 lines
#    # First line printed is where the signal found us
#  $stderr.print st.join("\n")
#  $stderr.print "\n"
#end 
##trap(:TSTP) { sigtstp }  # ^Z (Temporary stop) => show traceback

trap(:TSTP) { }  # Ignore ^Z (used to stop mk-hourly-sum2.rb)
  # NOTE: ^Z won't push owtmon.rb to background - ask it to stop with ^C
  #       or force it to stop with KILL -9 (after ps to get its procid)

$start_proc_time = Process::times;  $start_proc_date = Time.now

np_dist = nil

MkStrmSzDistribs = false  # Don't write stream-size distribs
szf = File.new('size.dat', 'w') if MkStrmSzDistribs

if Hours_reqd > 1  # Assume we want to process n hourly trace files
  uri = "pcapfile:#{trace_fn}"
  nu = 0;  other_uris = Array.new
  (1..Hours_reqd-1).each do
    new_uri = next_uri(uri)
    other_uris.push(new_uri)
    uri = new_uri
  end
  $owt.set_uris(other_uris)
end

n = 0
first_time = true
sum_start_ts = $owt.open
if LiveTrace
  log_fn = "#{file_dn(output_dn, sum_start_ts)}/#{utc_gts(sum_start_ts)}.fdf"
end

$owt.each do |msg|
  # print "#{Type_s[msg.type]} ---\n"
  case msg.type
  when OQ_TXTMSG
    type = msg.text[1..4]
    exit if msg.text =~ /Fatal/
    mts = msg.source  # g_last_ts for TXTMSGs
    case type  # txt message type
    when 'Time'  #Time
      $owt.log_msg(-1, msg.text)  # Write to log file
      $stdout.flush
      if MkStrmSzDistribs
        szf.print "#{utc_gts(mts)} "
        if first_time
          np_dist = Distribution.new(1, 400, 100, DT_LOG, 1.0)
          np_dist.write_x(szf, false)
          first_time = false
        else
          $owt.each_stream_size do |sz|
            np_dist.count(sz)
          end
          np_dist.write_y(szf, false)
          np_dist = Distribution.new(1, 400, 100, DT_LOG, 1.0)
        end
      end
    when 'Reco'  # Record: time to write sum2 file
      fn = "#{file_dn(output_dn, sum_start_ts)}" +
        "/#{utc_gts(sum_start_ts)}.#{SumFileType}"
      if distribs_reqd.nil?  # Not writing distribution info file
        write_sum_file(fn, sum_start_ts, mts) 
      end
      if day_change(mts)  # Start new log file
        log_fn = "#{file_dn(output_dn, mts)}/#{utc_gts(mts)}.fdf"
        $owt.new_log_file(log_fn)
      end
      sum_start_ts = mts
    when 'Stat', '#One'  # Stats record, startup message
      $owt.log_msg(-1, msg.text)  # Write to log file
      $stdout.flush
    when 'Shut'  #Shutdown
      $owt.log_msg(-1, msg.text)  # Write to log file
      $stdout.flush  # and console
      if MkStrmSzDistribs
        szf.print "#{utc_gts(mts)} "
        $owt.each_stream_size do |sz|
          np_dist.count(sz)
        end
        np_dist.write_y(szf, false)
      end
      fn = "#{file_dn(output_dn, sum_start_ts)}" +
        "/#{utc_gts(sum_start_ts)}.#{SumFileType}"
      write_sum_file(fn, sum_start_ts, mts)
        # $owt.each terminates after the #Shutdown message !?
    else  # Shouldn't happen
      $owt.log_msg(-1, "#{msg.text} <<<<<")  # Write to log file
      $stdout.flush  # Always written to console
    end
  when OQ_END_OF_FILE, OQ_TIMED_OUT  # Don't expect to see these in v2.0
    print "#{Type_s[msg.type]} ---\n"
    event = msg.source
  else  # Shouldn't happen !
    event = msg.source
    $owt.log_msg(0, "#{Type_s[msg.type]}:  ")
    first_fwd, src_addr, tot_pkts, lifetime = event.usage
    inward = first_fwd ? '->' : '<-'
    $owt.log_msg(-1, sprint("   %s %s  %d  %.6f *****", src_addr, inward, tot_pkts, lifetime))
    $stdout.flush
    exit
  end
#  n += 1
#    exit if n == 5
end

#szf.close if MkStrmSzDistribs

$fdf.close
