#!/usr/local/ruby

# 1405, Fri 22 Apr 11 (NZST) - from odin
#
# owt-support.rb: constants and utility methods for owtmon
#
# iatmon/owtmon: One-Way Traffic Monitor testbed
# Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

S_unclassified          =  0

S_tcp_probe             =  1
S_tcp_vertical_scan     =  2
S_tcp_horizontal_scan   =  3
S_tcp_unknown           =  4

S_udp_probe             =  5
S_udp_vertical_scan     =  6
S_udp_horizontal_scan   =  7
S_udp_unknown           =  8

S_icmp_only             =  9
S_backscatter           = 10
S_tcp_and_udp           = 11

S_utorrent              = 12
S_conficker_c           = 13
S_ipsec                 = 14

SrcTypeNames = [ 'Unclassified',  # 0
  'TCP probe', 'TCP vertical scan', 'TCP horizontal scan', 'TCP unknown',  # 1,2,3,4
  'UDP probe', 'UDP vertical scan', 'UDP horizontal scan', 'UDP unknown',  # 5,6,7,8
  'ICMP only', 'Backscatter', 'TCP and UDP',  # 9, 10, 11
  'uTorrent', 'Conficker P2P', 'IPSEC' ]  # 12, 13, 14

  
IAT_Other            = 0  # group numbers from IATdistribsFile.rb

IAT_short_lived      = 1
IAT_high_rate        = 2
IAT_DOS              = 3

IAT_3_left           = 4
IAT_3_even           = 5
IAT_3_right          = 6

IAT_stealth_3s_mode  = 7
IAT_stealth_spikes   = 8
IAT_stealth          = 9

GroupNames = [ 'Other', 'Short lived', 'High rate', 'DOS',
  '3s_mode left', '3s mode even', '3s mode right',
  'Stealth 3s mode', 'Stealth spikes', 'Stealth' ]


def run_command(cmd, print_output=false)  # return cmd's *last* output line
  $stdout.sync = true
  ll = ''
  IO.popen(cmd) do |f|
    until f.eof?
      ll = f.gets
      puts(ll) if print_output
    end
  end
  ll.chomp
end

def already_there(fn)
  return false if !File.exists?(fn)
  return false if File.size?(fn).nil?  # nil if size zero
  true
end

def check_dir_exists(dn)
  e = File.exists?(dn); d = File.directory?(dn)
  if e && !d
    print "#{dn} is an existing file, not creating directory !!!\n"
    exit
  end
  Dir.mkdir(dn) if !e
end

# Telescope times are UTC, but Ruby formats them in local time!

def utc_gts(local_ts)  # Convert ts to gnuplot timefmt
  ds = local_ts.to_i+20  # strfmtime will round down to closest hour
  Time.at(ds).gmtime.strftime("%Y%m%d.%H%M")
    # Assume that gmtime uses TZ environment variable
    # time format suitable for gnuplot !
end

def utc_hts(local_ts)  # Convert ts to ymd date with hyphens
  ds = local_ts.to_i+20  # strfmtime will round down to closest hour
  Time.at(ds).gmtime.strftime("%Y-%m-%d")
end

class SummaryFile
  def find_section(sect)
    if sect.size == 4  # Old section names
      prefix = ('#' + sect)[0..4]
    else
      prefix = '#' + sect
    end
    begin  # May reach EOF
      @sf.each do |line|
        ix = line.index(prefix) 
        if !ix.nil? && ix == 0
          la = line.split(' ', 2)  # Always makes two strings!
          return la[1].size == 0 ? nil : la[1]
        end
      end
    rescue EOFError
      return nil
    end
    return nil
  end

  def initialize(sf_name, pr=true)
    @n_types = SrcTypeNames.size
    @n_groups = GroupNames.size
    @sf_name = sf_name
    @sf = File.open(@sf_name, 'r')  # Open the summary file
    @pr = pr

    find_section('version')
    @format_ver, @owtmon_ver, @c_meter_ver = @sf.readline.split(' ')

    find_section('meter_info')
    @meter_loc, @ymdh, @tf_sz, @minpkts, tw = @sf.readline.split(' ')
    # @tf_sz is size in GB for a trace file, hostname for a live interface
    @two_way = tw == 'two-way'
    print "#{@meter_loc}, #{@ymdh}, #{@tf_sz}, #{@minpkts}" \
      + ", two-way=#{@two_way}\n" if @pr
  end

  def readreals(msg)
    la = @sf.readline.split
    return nil if la.size == 0  #  Blank line
    if la.size != @n_groups
      $stdout.print "#{la.size} group values read, expected #{@n_groups}\n"
      return nil
    end
    va = Array.new(la.size)
    la.each_index { |x| va[x] = la[x].to_f }
    return va
  end

  def get_counts(prefix)
    if find_section(prefix).nil?  # source_counts
      print "#{@ymdh} file has no #sour record <<<\n"
      return nil
    end
    xt_counts = Array.new
    begin
      r = readreals('IAT_groups')
      xt_counts.push(r) if !r.nil?
    end until r.nil?
    if xt_counts.size != @n_types
      print "#{prefix}: #{line.sise} values read, expected #{n_expected}\n"
      return nil
    end
    xt_counts
  end

  def close
    @sf.close
  end

  def reopen
    @sf = File.open(@sf_name, 'r')
  end

  attr_reader :format_ver, :meter_loc, :ymdh, :two_way, :sf_name
end  # SummaryFile

class SummaryDataSet
  def initialize(sum_dir, start_date,last_date, pr=true)
    @sum_dir = sum_dir; 
    @start_date = start_date;  @last_date = last_date
    @pr = pr
  end

  def each
  print "scan_summaries: sum_dir=#{@sum_dir}<\n"
    # Expects an associated block
    Dir.glob("#{@sum_dir}/????-??-??").sort.each do |an|
      date = File.basename(an)
#    print "  date=#{date}, date<start_date=#{date < start_date}\n"
      next if date < @start_date
      break if date > @last_date
#    print "#{date}\n"
      Dir.glob("#{@sum_dir}/#{date}/*.#{SumFileType}").sort.each do |fulln|
        sf = SummaryFile.new(fulln, @pr)
        yield sf
        sf.close
        # exit  # Just do one file!
      end
    end
  end
end  # SummaryDataSet
